/*-
 * Copyright (c) 2006 Nate Lawson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/select.h>
#include <sys/socket.h>
#ifdef __FreeBSD__
#include <sys/endian.h>
#else
#define	bswap32(x)			\
	((((x) & 0xff000000) >> 24) |	\
	 (((x) & 0x00ff0000) >>  8) |	\
	 (((x) & 0x0000ff00) <<  8) |	\
	 (((x) & 0x000000ff) << 24))
#define	bswap16(x)			\
	((((x) & 0xff00) >> 8) |	\
	 (((x) & 0x00ff) << 8))
#endif /* __FreeBSD__ */

static int localEndian = BYTE_ORDER;
static int remoteEndian = BYTE_ORDER;

/* Given a string, find the end quote, skipping escaped quotes. */
static char *
match_quote(char *str, char *end)
{
	while (str < end) {
		if (*str == '\\')
			str++;
		else if (*str == '"')
			break;
		str++;
	}

	if (*str != '"')
		str = NULL;
	return (str);
}

static int
trans_string(char *in, int len)
{
	char *end, *out;

	out = in;
	for (end = in + len; in < end; in++, out++) {
		if (*in != '\\') {
			*out = *in;
			continue;
		}

		switch (in[1]) {
		case 'n':
			*out = '\n';
			break;
		case 'r':
			*out = '\r';
			break;
		case 't':
			*out = '\t';
			break;
		default:
			*out = in[1];
			break;
		}

		in++;
		len--;
	}

	return (len);
}

static int
asc2bytes(int s, char *buf, int len)
{
	char *ptr, *end, *next;
	int diff, str_len;
	u_long val;
	uint16_t half;
	uint8_t byte;

	ptr = buf;
	end = buf + len;
	while (ptr < end) {
		if (*ptr == '\\') {
			ptr++;
			write(s, ptr, 1);
			ptr++;
		} else if (*ptr == '"') {
			/* Output a translated string. */
			ptr++;
			next = match_quote(ptr, end);
			if (next == NULL) {
				fprintf(stderr,
				    "syntax error: non-matched quotes\n");
				return (-1);
			}
			str_len = trans_string(ptr, next - ptr);
			write(s, ptr, str_len);
			ptr = next + 1;
		} else if (isxdigit(*ptr) || (*ptr == '0' && ptr[1] == 'x')) {
			/*
			 * Translate hex values to binary, keeping word
			 * groupings and making endianness swaps as
			 * necessary.
			 */
			if (*ptr == '0' && ptr[1] == 'x')
				ptr += 2;
			val = strtoul(ptr, &next, 16);
			diff = next - ptr;
			if (diff == 0) {
				printf("invalid hex: %s\n", ptr);
			} else if (diff <= 2) {
				byte = (uint8_t)val;
				write(s, &byte, sizeof(byte));
			} else if (diff <= 4) {
				half = (uint16_t)val;
				if (localEndian != remoteEndian)
					half = bswap16(half);
				write(s, &half, sizeof(half));
			} else {
				if (localEndian != remoteEndian)
					val = bswap32(val);
				write(s, &val, sizeof(val));
			}
			ptr = next;
		} else {
			/* Suppress any other characters. */
			ptr++;
		}
	}

	return (0);
}

static int
bytes2asc(int s, uint8_t *buf, int len)
{
	int i;

	if (len == 0)
		return (0);

	printf("#   ");
	for (i = 0; i < len; i++)
		printf("%02x ", buf[i]);
	printf("\n");

	return (0);
}

int
main(int argc, char **argv)
{
	int socks[2], nfds, err, len, remfd;
	fd_set saved_fds, fds;
	char buf[256], *arg;

	if (argc < 2) {
		fprintf(stderr,
		    "usage: %s prog [-eb|-el] [args...]\n", argv[0]);
		exit(1);
	}
	argv++;

	/* Handle the -e arg.  Probably should use getopt(3) instead. */
	arg = *argv;
	if (*arg++ == '-') {
		switch (*arg++) {
		case 'e':
			if (tolower(*arg) == 'b')
				remoteEndian = BIG_ENDIAN;
			else if (tolower(*arg) == 'l')
				remoteEndian = LITTLE_ENDIAN;
			else {
				fprintf(stderr,
				    "arg must be [b]ig or [l]ittle endian\n");
				exit(1);
			}
			break;
		default:
			fprintf(stderr, "unknown flag: %s\n", argv[0]);
			exit(1);
		}
		argv++;
	}

	/* Get the sockets and start the child process. */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, socks) != 0) {
		perror("socketpair");
		exit(1);
	}
	switch (fork()) {
	case 0:
		/* Hook up the child process to the local socket. */
		dup2(socks[1], STDIN_FILENO);
		dup2(socks[1], STDOUT_FILENO);
		close(socks[0]);
		execvp(*argv, argv);
		perror(*argv);
		exit(1);
	case -1:
		perror("fork");
		exit(1);
	default:
		/* Prepare the parent side of the socket. */
		remfd = socks[0];
		close(socks[1]);
		break;
	}

	/* Set up our select args in advance. */
	FD_ZERO(&saved_fds);
	FD_SET(remfd, &saved_fds);
	FD_SET(STDIN_FILENO, &saved_fds);
	nfds = remfd + 1;

	/* Handle upstream and downstream I/O independently. */
	for (;;) {
		fds = saved_fds;
		err = select(nfds, &fds, NULL, NULL, NULL);
		if (err < 0) {
			perror("select");
			exit(1);
		}

		/* Got input from upstream, translate to binary and send. */
		if (FD_ISSET(STDIN_FILENO, &fds)) {
			len = read(STDIN_FILENO, buf, sizeof(buf));
			if (len <= 0)
				break;
			asc2bytes(remfd, buf, len);
		}

		/* Got output from downstream, translate to ASCII and send. */
		if (FD_ISSET(remfd, &fds)) {
			len = read(remfd, buf, sizeof(buf));
			if (len <= 0)
				break;
			bytes2asc(STDOUT_FILENO, (uint8_t *)buf, len);
		}
	}

	exit(0);
}
